@@ -1,5 +1,3 @@
-#!/usr/local/bin/ruby -w
-
 # question.rb
 #
 #  Created by James Edward Gray II on 2005-04-26.
@@ -33,9 +31,9 @@ class HighLine
     #
     def initialize( question, answer_type )
       # initialize instance data
-      @question    = question
+      @question    = question.dup
       @answer_type = answer_type
-      
+
       @character    = nil
       @limit        = nil
       @echo         = true
@@ -49,19 +47,20 @@ class HighLine
       @in           = nil
       @confirm      = nil
       @gather       = false
+      @verify_match = false
       @first_answer = nil
       @directory    = Pathname.new(File.expand_path(File.dirname($0)))
       @glob         = "*"
       @responses    = Hash.new
       @overwrite    = false
-      
+
       # allow block to override settings
       yield self if block_given?
 
       # finalize responses based on settings
       build_responses
     end
-    
+
     # The ERb template of the question to be asked.
     attr_accessor :question
     # The type that will be used to convert this answer.
@@ -152,6 +151,12 @@ class HighLine
     # 
     attr_accessor :gather
     # 
+    # When set to +true+ multiple entries will be collected according to
+    # the setting for _gather_, except they will be required to match
+    # each other. Multiple identical entries will return a single answer.
+    #
+    attr_accessor :verify_match
+    #
     # When set to a non *nil* value, this will be tried as an answer to the
     # question.  If this answer passes validations, it will become the result
     # without the user ever being prompted.  Otherwise this value is discarded, 
@@ -236,6 +241,8 @@ class HighLine
                      :not_in_range         =>
                        "Your answer isn't within the expected range " +
                        "(#{expected_range}).",
+                     :mismatch             =>
+                       "Your entries didn't match.",
                      :not_valid            =>
                        "Your answer isn't valid (must match " +
                        "#{@validate.inspect})." }.merge(@responses)
@@ -289,6 +296,7 @@ class HighLine
     # Pathname::               Same as File, save that a Pathname object is
     #                          returned.
     # String::                 Answer is converted with Kernel.String().
+    # HighLine::String::       Answer is converted with HighLine::String()
     # Regexp::                 Answer is fed to Regexp.new().
     # Symbol::                 The method to_sym() is called on answer and
     #                          the result returned.
@@ -301,6 +309,8 @@ class HighLine
     def convert( answer_string )
       if @answer_type.nil?
         answer_string
+      elsif [::String, HighLine::String].include?(@answer_type)
+        HighLine::String(answer_string)
       elsif [Float, Integer, String].include?(@answer_type)
         Kernel.send(@answer_type.to_s.to_sym, answer_string)
       elsif @answer_type == Symbol
@@ -329,7 +339,7 @@ class HighLine
       end
     end
 
-    # Returns a english explination of the current range settings.
+    # Returns an English explanation of the current range settings.
     def expected_range(  )
       expected = [ ]
 
